r"""Contain utility functions to manipulate null values in
DataFrames."""

from __future__ import annotations

__all__ = ["propagate_nulls"]

import polars as pl


def propagate_nulls(frame: pl.DataFrame, frame_with_null: pl.DataFrame) -> pl.DataFrame:
    r"""Propagate the null values from ``frame_with_null`` to ``frame``.

    Args:
        frame: The input DataFrame where to add ``None`` values based
            on ``frame_with_null``.
        frame_with_null: The DataFrame with the ``None`` values to
            propagate to ``frame``.

    Returns:
        The output DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.null import propagate_nulls
    >>> frame_with_null = pl.DataFrame(
    ...     {
    ...         "col1": [1, None, 3, float("nan"), 5],
    ...         "col2": ["1", "2", None, "4", "5"],
    ...         "col3": [10, 20, 30, None, 50],
    ...     },
    ...     schema={"col1": pl.Float32, "col2": pl.String, "col3": pl.Int64},
    ... )
    >>> frame = frame_with_null.fill_null(99).fill_nan(99)
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ 1.0  ┆ 1    ┆ 10   │
    │ 99.0 ┆ 2    ┆ 20   │
    │ 3.0  ┆ null ┆ 30   │
    │ 99.0 ┆ 4    ┆ 99   │
    │ 5.0  ┆ 5    ┆ 50   │
    └──────┴──────┴──────┘
    >>> out = propagate_nulls(frame=frame, frame_with_null=frame_with_null)
    >>> out
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ 1.0  ┆ 1    ┆ 10   │
    │ null ┆ 2    ┆ 20   │
    │ 3.0  ┆ null ┆ 30   │
    │ 99.0 ┆ 4    ┆ null │
    │ 5.0  ┆ 5    ┆ 50   │
    └──────┴──────┴──────┘

    ```
    """
    columns = frame.columns
    return (
        frame.with_columns(frame_with_null.select(pl.all().is_null().name.suffix("__@@isnull@@_")))
        .with_columns(
            pl.when(~pl.col(col + "__@@isnull@@_")).then(pl.col(col)).otherwise(None)
            for col in columns
        )
        .select(columns)
    )



r"""Contain the base class to implement a DataFrame exporter."""

from __future__ import annotations

__all__ = ["BaseExporter", "is_exporter_config", "setup_exporter"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class BaseExporter(ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a DataFrame exporter.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.exporter import ParquetExporter
    >>> exporter = ParquetExporter(path="/path/to/frame.parquet")
    >>> exporter
    ParquetExporter(path=/path/to/frame.parquet)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> exporter.export(frame)  # doctest: +SKIP

    ```
    """

    @abstractmethod
    def export(self, frame: pl.DataFrame) -> None:
        r"""Export a DataFrame.

        Args:
            frame: The DataFrame to export.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from grizz.exporter import ParquetExporter
        >>> exporter = ParquetExporter(path="/path/to/frame.parquet")
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> exporter.export(frame)  # doctest: +SKIP

        ```
        """


def is_exporter_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseExporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseExporter`` object.

    Example usage:

    ```pycon

    >>> from grizz.exporter import is_exporter_config
    >>> is_exporter_config(
    ...     {"_target_": "grizz.exporter.ParquetExporter", "path": "/path/to/data.parquet"}
    ... )
    True

    ```
    """
    return is_object_config(config, BaseExporter)


def setup_exporter(
    exporter: BaseExporter | dict,
) -> BaseExporter:
    r"""Set up an exporter.

    The exporter is instantiated from its configuration
    by using the ``BaseExporter`` factory function.

    Args:
        exporter: A exporter or its configuration.

    Returns:
        An instantiated exporter.

    Example usage:

    ```pycon

    >>> from grizz.exporter import setup_exporter
    >>> exporter = setup_exporter(
    ...     {"_target_": "grizz.exporter.ParquetExporter", "path": "/path/to/data.parquet"}
    ... )
    >>> exporter
    ParquetExporter(path=/path/to/data.parquet)

    ```
    """
    if isinstance(exporter, dict):
        logger.info("Initializing an exporter from its configuration... ")
        exporter = BaseExporter.factory(**exporter)
    if not isinstance(exporter, BaseExporter):
        logger.warning(f"exporter is not a `BaseExporter` (received: {type(exporter)})")
    return exporter
